C2_RemokonResponder {

	// 38nonprivate@durftal.com
	
	var responder, enabled = false;
	var actionRefArray, actionArray, actionDict, deviceDict;
	var screenDict, widgetDict;		// match up names of screens and widgets to indexes into actionRefArray
	var addr, grid;
	// grid - instances of C2_akaGrid

	*new {
	
		^super.new.init;
	
	}
	
	init {
		
		// action reference array
		// multi dimensional array to store and lookup actions by: deviceID, screen, widgetID
		// returns index into actionArray, where the actual actions are stored
		// and also the current values - might be useful at some point
		actionRefArray = [];
		
		// storing and retrieving the actions themselves - 
		// actions and widgets are decoupled
		// maybe 2 widgets can control the same thing
		// actions are stored by name in the actionDict
		// which returns the index of the action itself in the actionArray
		
		actionArray = [];					// stores the actions
		actionDict = Dictionary.new;		// action name key -> actionArray indexes
		
		deviceDict = Dictionary.new;		// deviceID - > actionRefArray 1st index
		
		screenDict = 
			Dictionary[
				"mix" -> 0, "pad" -> 1, "xy" -> 2, "grid" -> 3
			];
		
		widgetDict = 
			Array[
			
				// "mix"
				Dictionary[
					"button1" -> 0, "button2" -> 1, "button3" -> 2, "button4" -> 3,
					"slider1" -> 10, "slider2" -> 11, "slider3" -> 12, "slider4" -> 13,
					"xfader" -> 20
				],
				
				// "pad"
				Dictionary[
					"button1" -> 0, "button2" -> 1, "button3" -> 2, "button4" -> 3,
					"pad1" -> 10, "pad2" -> 11, "pad3" -> 12, "pad4" -> 13,
					"pad5" -> 14, "pad6" -> 15, "pad7" -> 16, "pad8" -> 17,
					"pad9" -> 18, "pad10" -> 19, "pad11" -> 20, "pad12" -> 21,
					"pad13" -> 22, "pad14" -> 23, "pad15" -> 24, "pad16" -> 25				],
					
				// "xy"
				Dictionary[
					"button1" -> 0, "button2" -> 1, "button3" -> 2, "button4" -> 3,
					"xy" -> 10
				],
				
				// "grid" - btw the actual grid bits are handled with the C2_akaGrid class
				Dictionary[
					"button1" -> 0, "button2" -> 1, "button3" -> 2, "button4" -> 3
				]	
			
			
			];
			
		for(10, 73, 
			{
			arg c;
			widgetDict[3].put("grid" ++ (c - 9).asString, c);
			}
		);
			
		addr = []; grid = [];
			
	}
	
	addDevice {
	
		arg deviceID, hostname, port;
		var actionRefArrayIndex, baseNo, row, column;
		
		actionRefArray = actionRefArray.add(Array.newClear(4));
		actionRefArrayIndex = actionRefArray.size - 1;
		
		hostname = hostname ? "127.0.0.1"; port = port ? 8000;
		addr = addr.add(NetAddr(hostname, port));
		grid = grid.add(
		
			C2_Grid4Remokon.new(
			
				noOfColumns: 8, noOfRows: 8, 
				noOfStates: 2, idNo: deviceID, prefix: "", hostname: hostname, port: port
			
			)
		
		);
		
		for(10, 73,
			{
			arg c;
			
			baseNo = c - 10;
			row = baseNo.div(8);
			column = baseNo.mod(8);
			
			grid[actionRefArrayIndex].registerFunction(
				key: c.asString,
				column: column, row: row, width: 1, height: 1,
				buttonResponse: \noChange,
				function: {
					arg val1, status;
					this.handleResponse(
						deviceID: actionRefArrayIndex, screenID: 3, 
						widgetID: c, widgetType: nil, 
						status: status, val1: val1
					);
		
				}
			
			);
			
			
			}
		);
		
		grid[actionRefArrayIndex].refreshDur_(0.002);
		grid[actionRefArrayIndex].enable_noResponder;

		
		
		// there will always be some "dummy" gaps in these arrays, as widget id numbers aren't always sequential
		// "mix"
		actionRefArray[actionRefArrayIndex][0] = Array.newClear(21);


		// "pad"
		actionRefArray[actionRefArrayIndex][1] = Array.newClear(26);
		// "xy"
		actionRefArray[actionRefArrayIndex][2] = Array.newClear(11);
		// "grid"
		actionRefArray[actionRefArrayIndex][3] = Array.newClear(74);
		
		// initial values may get referred to before they are changed or set
		for(0, 3, 
			{
			arg c;
				
			actionRefArray[actionRefArrayIndex][c].do({
				|item, index|
				// action ref index (index into action ref array), initial value1, initial value 2
				actionRefArray[actionRefArrayIndex][c][index] = [nil, 0, 0];
				
			});
			
			}
		);
		
		deviceDict.put(deviceID, actionRefArrayIndex);
		
	}
	
	changeNetAddrByDeviceID {
	
		arg deviceID, hostname, port;
		var deviceIndex;
		
		deviceIndex = deviceDict[deviceID];
		hostname = hostname ? "127.0.0.1"; port = port ? 8000;
		
		if(deviceIndex != nil,
			{
			addr[deviceIndex] = NetAddr(hostname, port);
			grid[deviceIndex].hostname_(hostname).port_(port);
			}
		);
		
	
	
	}
	
	addActionByName {
	
		arg deviceID, screenName, widgetName, action;
		var screenID, widgetID;
		
	//	"addActionByName".postln;
		
		// check that the names given are valid
		screenID = screenDict[screenName];
		if(screenID != nil, {widgetID = widgetDict[screenID][widgetName]});
		if(widgetID != nil, 
			{
			this.addActionByID( deviceID, screenID, widgetID, action)
			}
		);
	
	}
	
	// best not to use directly
	addActionByID {
	
		arg deviceID, screenID, widgetID, action;
		var deviceIndex;
				
		deviceIndex = deviceDict[deviceID];
		actionArray = actionArray.add(action);
		actionRefArray[deviceIndex][screenID][widgetID][0] = actionArray.size - 1;
		// [action index, current value 1, current value 2]
	
	}
	
	setValueByName {
	
		arg deviceID, screenName, widgetName, val1, val2;
		var screenID, widgetID;
				
		// check that the names given are valid
		screenID = screenDict[screenName];
		if(screenID != nil, {widgetID = widgetDict[screenID][widgetName]});
		if(widgetID != nil, {this.setValueByID( deviceID, screenID, widgetID, val1, val2)});
	
	}
	
	// best not to use directly
	setValueByID {
		
		arg deviceID, screenID, widgetID, val1, val2;
		var deviceIndex;
				
		deviceIndex = deviceDict[deviceID];
		if(addr[deviceIndex].class != NetAddr, {^nil});
	//	["/set", addr[deviceIndex], deviceID, screenID, widgetID, 1, 1, val1, val2 ].postln;
		addr[deviceIndex].sendMsg("/set", deviceID, screenID, widgetID, 1, 1, val1, val2 );
		
	
		if( actionRefArray[deviceIndex][screenID][widgetID] != nil,
			{	
			actionRefArray[deviceIndex][screenID][widgetID][1] = val1;
			actionRefArray[deviceIndex][screenID][widgetID][2] = val2;
			}
		);
		
	}
	
	getValueByName {
	
		arg deviceID, screenName, widgetName;
		var screenID, widgetID, out;
		
		// check that the names given are valid
		screenID = screenDict[screenName];
		if(screenID != nil, {widgetID = widgetDict[screenID][widgetName]});
		if(widgetID != nil, {out = this.getValueByID( deviceID, screenID, widgetID)});
		
		^out;
		
	}
	
	getValueByID {
	
		arg deviceID, screenID, widgetID;
		var deviceIndex;
		
		deviceIndex = deviceDict[deviceID];
		
		// return array in this format: [val1, val2]
		^actionRefArray[deviceIndex][screenID][widgetID].copyRange(1, 2);
		
	}
	
	enable {
	
		enabled = true;
		
		if(responder != nil, {responder.remove});

		responder = OSCresponder(nil, "/event", 
			{ 
			arg time, responder, msg, addr;	 
			var deviceID, screenID, widgetID, widgetType, status, val1, val2;
			
			//[time, responder.addr, msg, addr].postln;
			
			
			// status: down (1), continue (2) or up (0)
			deviceID = msg[1]; screenID = msg[2]; widgetID = msg[3]; widgetType = msg[4];
			status = msg[5]; val1 = msg[6]; val2 = msg[7];
			
			if(screenID < 4,
				{
				this.handleResponse(time, deviceID, screenID, widgetID, widgetType, status, val1, val2);
				}
			);
			
		     }
		).add;	
	
	}
	
	
	disable {
	
		enabled = false;
		if(responder != nil, {responder.remove});
	
	}
	
	
	getData {
	
	}
	
	setData {
	
	}
	
	handleResponse {
	
		arg time, deviceID, screenID, widgetID, widgetType, status, val1, val2;
		var deviceIndex;
		
		if(widgetID == -1, {^nil});		// the screen got changed, that's all
		
		deviceIndex = deviceDict[deviceID];
		
		if(deviceDict[deviceID] != nil,
			{
			if( (screenID == 3) and: (widgetID >= 10),
				{
				grid[deviceIndex].handleEventByID(time, widgetID, status);
				}
			);
			
			if(status >= 1,
				{
				actionRefArray[deviceIndex][screenID][widgetID][1] = val1;
				actionRefArray[deviceIndex][screenID][widgetID][2] = val2;
				}
			);

			if( actionRefArray[deviceIndex][screenID][widgetID][0] != nil,
				{
				actionArray[actionRefArray[deviceIndex][screenID][widgetID][0]].value(status, val1, val2);
				
				}
			);
			
			}
		);
	
	
	}
	

}

C2_Grid4Remokon : C2_OSCGridInterface_4Remokon {


	enable {
	
		enabled = true;
		
		if(responder != nil, {responder.remove});

		responder = OSCresponder(nil, "/event", 
			{ 
			arg time, responder, msg;	  
			var column, row, status, baseNo;  
			
		//	msg.postln;
			
			if( msg[2] == 3,
				{
				baseNo = msg[3] - 10;
				row = baseNo.div(noOfRows);
				column = baseNo.mod(noOfRows);
				status = msg[5];
				this.handleResponse(column, row, status);
				}
			);

		        
		}).add;	
		
		this.refreshAll;
	
	}
	
	enable_noResponder {
	
		enabled = true;
		this.refreshAll;
	
	}
	
	handleEvent {
	
		arg time, msg;
		var column, row, status, baseNo; 
		
		if( msg[2] == 3,
			{
			baseNo = msg[3] - 10;
			row = baseNo.div(noOfRows);
			column = baseNo.mod(noOfRows);
			status = msg[5];
			this.handleResponse(column, row, status);
			}
		);	
	
	}
	
	// akaGrid only!
	handleEventByID {
	
		arg time, widgetID, status;
		var column, row, baseNo; 
		
		baseNo = widgetID - 10;
		row = baseNo.div(noOfRows);
		column = baseNo.mod(noOfRows);

		this.handleResponse(column, row, status);
	
	}
	
	// PRIVATE... DO NOT USE DIRECTLY!
	led_private {
	
		arg column, row, state;
		var ledNo;
		
		ledNo = 10 + (row * 8) + column;
		addr.sendMsg( "/set", idNo, 3, ledNo, 1, 1, state);

	
	}


	

}


C2_OSCGridInterface_4Remokon {

	var <prefix, <hostname, responder, <hostPort, <listenPort;
	var <noOfColumns, <noOfRows, <idNo, <addr, <port;
	var addr;
	var noOfStates;
	var blinkRoutine, whichBlinkRoutine = 0, blinkDur = 0.012, blinkArray, blinkStatus = 0;
	var stateArray, actionRefArray, actionArray, actionDict, responseArray;
	var valueArray;
	var actionObject;
	var <enabled = false;
	var <>refreshDur = 0;
	var <alphaDisplayFlag = 0, alphaDisplayBounds, alphaDisplayString, alphaDisplayDict, alphaDisplayMap;
	var alphaDisplayStart;		// [column, row]
	var pressBag;			// keeps track of co-ordinates of all currently pressed buttons (status 1)
						// in order to make multiple press combinations work
						
	var pressTime;
	// an array with empty slots for each location
	// give the Process.elapsed time for each on press (status = 1)
	// use for double trigger prevention handling ( a record of when the last press was)
	
	var <>doubleTriggerBuffer = 0.03;
	// time in seconds, within which, having received a press on (status 1) off (0) then on again (1), the off is ignored
	// if this time lapses after the 0 and a 1 isn't received again, then it's definately registered as an off
	
	var <>allAction;
	
	var overlayStyle;
	var negButtonDisplayDict;
	var flashBag;
	
	var currentActualStateArray;
	
	var <>virtualDisplayAddr, <>virtualDisplayFlag = 1;

	*new {
	
		arg noOfColumns, noOfRows, noOfStates, idNo, prefix, hostname, port, specialParams;
		
		noOfColumns = noOfColumns ? 16; noOfRows = noOfRows ? 8; noOfStates = noOfStates ? 3; 
		idNo = idNo ? 0; prefix = prefix ? ""; 
		hostname = hostname ? "127.0.0.1"; port = port ? 8080;
	
		^super.new.init(noOfColumns, noOfRows, noOfStates, idNo, prefix, hostname, port, specialParams);
	
	}
	
	
	init {
	
		arg noOfColumnsArg, noOfRowsArg, noOfStatesArg, idNoArg, prefixArg, hostnameArg, portArg, specialParams;
		
		noOfColumns = noOfColumnsArg; noOfRows = noOfRowsArg;
		noOfStates = noOfStatesArg; idNo = idNoArg; prefix = prefixArg;
		hostname = hostnameArg; port = portArg;
		
		
		addr = NetAddr(hostname, port); 
		this.initSpecial(specialParams);
		
		virtualDisplayAddr = NetAddr("192.168.2.19", 57120);
		
		allAction = {1};
		
		blinkArray = [];
		flashBag = Bag.new;
	
		// manually set led state (overridden by actions)
		stateArray = Array.fill2D( noOfColumns, noOfRows, 0 );
		currentActualStateArray = Array.fill2D( noOfColumns, noOfRows, 0 );
		
		// key name reference of each led to look up in actionDict
		// 2 layers - default goes to first layer
		// 1st layer changes status of objects etc. 2nd layer is always completely invisible
		// whatever the functions say
		actionRefArray = [Array.fill2D( noOfColumns, noOfRows, nil ), Array.fill2D( noOfColumns, noOfRows, nil )];
		actionArray = [];					// stores the actions
		actionDict = Dictionary.new;		// action name key -> actionArray index
		
		valueArray = Array.fill2D( noOfColumns, noOfRows, 0 );
		
		blinkRoutine = Array.newClear(8);
		blinkRoutine.do({
			|item, index|
			blinkRoutine[index] = 
				Routine({
				
				var flashStatus = 0, bc = 0;
				
					9999999.do({
					
						if(blinkArray.size > 0,
							{		
						//	index.postln;
							blinkArray.do({
								|item, index|
								this.led_private(item[0], item[1], 0);
							});
							
							blinkDur.wait;
								
							blinkArray.do({
								|item, index|
								this.led_private(item[0], item[1], 1);
							});
							
							}
						);
						
						if(bc.mod(7) == 0)
							{
						//	["ding", flashBag].postln;
							flashStatus = 1 - flashStatus;
							if(flashBag.size > 0)
								{
								flashBag.asArray.do({ |item, index|
									this.led_private(item[0], item[1], flashStatus);
								});
								
								};
								
							};
						
						blinkDur.wait;
						bc = bc + 1;
					// bc.postln;
					});
				
				
				});	
		});
		
		negButtonDisplayDict = 
			Dictionary[
				"+" -> [0, 1, 0, 1, 1, 1, 0, 1, 0],
				"-" -> [0, 0, 0, 1, 1, 1, 0, 0, 0]
			];
		
		alphaDisplayDict = 
			Dictionary[
				"a" -> [1, 0, 1, 1, 1, 1],
				"b" -> [1, 1, 1, 0, 1, 1],
				"c" -> [1, 1, 1, 1, 0, 1],
				"d" -> [0, 1, 1, 1, 1, 1],
				"e" -> [1, 0, 1, 1, 0, 1],
				"f" -> [1, 0, 1, 1, 0, 0],
				"g" -> [0, 1, 1, 0, 0, 1],
				"h" -> [1, 1, 1, 0, 1, 0],
				"i" -> [1, 0, 0, 0, 1, 1],
				"j" -> [0, 0, 1, 1, 1, 1],
				"k" -> [0, 1, 0, 1, 0, 1],
				"l" -> [1, 1, 1, 0, 0, 1],
				"m" -> [1, 0, 0, 0, 1, 0],
				"n" -> [0, 1, 0, 0, 0, 1],
				"o" -> [1, 1, 1, 1, 1, 1],
				"p" -> [1, 1, 1, 1, 1, 0],
				"q" -> [1, 1, 0, 1, 1, 1],
				"r" -> [1, 1, 1, 1, 0, 0],
				"s" -> [1, 1, 0, 0, 1, 1],
				"t" -> [1, 0, 0, 1, 1, 1],
				"u" -> [1, 1, 0, 0, 0, 1], 
				"v" -> [0, 0, 1, 1, 1, 0],
				"w" -> [0, 1, 0, 1, 0, 0],
				"x" -> [1, 0, 1, 0, 1, 0],
				"y" -> [0, 1, 1, 1, 1, 0],
				"z" -> [1, 0, 1, 1, 1, 0],
				
				"-" -> [0, 1, 0, 0, 1, 0],
				"+" -> [0, 1, 0, 1, 1, 1, 0, 1, 0],
				"_" -> [0, 0, 1, 0, 0, 1],
				"." -> [0, 0, 1],
				"'" -> [1, 1, 0],
				"|" -> [0, 1, 0],
				"!" -> [0, 0, 0, 0, 0, 0],
				
				"0" -> [1, 1, 1, 1, 0, 1, 1, 1, 1],
				"1" -> [1, 0, 1, 1, 1, 1, 0, 0, 1],
				"2" -> [0, 1, 1, 0, 1, 0, 1, 1, 0],
				"3" -> [1, 0, 1, 1, 1, 1, 1, 1, 1],
				"4" -> [1, 1, 0, 0, 1, 0, 1, 1, 1],
				"5" -> [1, 0, 1, 0, 1, 0, 1, 0, 1],
				"6" -> [1, 1, 1, 0, 1, 1, 0, 1, 1],
				"7" -> [1, 0, 0, 1, 0, 1, 1, 1, 0],
				"8" -> [1, 0, 1, 1, 1, 1, 1, 0, 1],
				"9" -> [1, 1, 0, 1, 1, 0, 1, 1, 1]

			];
		
		alphaDisplayBounds = [0, 0, 15, 7];
		alphaDisplayString = "";
		
		pressBag = Bag.new;
		pressTime = Array.fill2D( noOfColumns, noOfRows, 0 );
	
		
		
	}

	initSpecial {
	
		arg specialParams;
	
		// ie for monome:
		// arg cable, intensity
		// etc
		
		// cable = cable ? "top";
		// intensity = intensity ? 0.5;

		
//		addr.sendMsg("/sys/prefix", unitNo, prefix);
//		addr.sendMsg("/sys/cable", unitNo, cable);
//		addr.sendMsg("/sys/intensity", unitNo, intensity);
	
	}
	
	
	enable {
	
		enabled = true;
		
		if(responder != nil, {responder.remove});
		
		// the following would work for a monome
		// but as this is the daddy class, it shouldn't be used directly
		// and this method should be replaced in classes that inherit from here
		// it's left here as an example
		
//		responder = OSCresponder(nil, prefix ++ "/press", 
//			{ 
//			arg time, responder, msg;	  
//			var column, row, status;   
//			column = msg[0]; row = msg[1]; status	 = msg[2];
//			this.handleResponse(column, row, status);
//		        
//		}).add;	
		
		this.clearAll;
		this.refreshAll;
	
	}
	
	handleEvent { 
	
	// receive events by other means than responder
		arg time, msg;
		
	
	}
	
	disable {
	
		enabled = false;
		this.startStopBlink(0);
		if(responder != nil, {responder.remove});
	
	}
	
	led_set {
	
	// set single led state without refreshing
	// handy for building up states pixel by pixel, then just have one refreshAll at the end

		arg column = 0, row = 0, state = 0;
		var detectIndex;
		
		column = column.clip(0, noOfColumns - 1); row = row.clip(0, noOfRows - 1);
		state = state.clip(0, noOfStates - 1);
		
		stateArray[column][row] = state;
	
	
	}
	
	
	led {
	
		arg column = 0, row = 0, state = 0, changeBlink = false;
	//	var detectIndex;
		
		column = column.clip(0, noOfColumns - 1); row = row.clip(0, noOfRows - 1);
		state = state.clip(0, noOfStates - 1);
		
		stateArray[column][row] = state;
	
		if(enabled == true, 
			{
			this.refreshArea(column, column, row, row, changeBlink);
			}
		);

	
	}
	
	led_phantom { arg column, row, state1, state2, dur = 0.02;

		if(enabled == true, 
			{	
			AppClock.play(Routine({
				this.led_private(column, row, state1);
				dur.wait;
				this.led_private(column, row, state2);
			}));
			
			}
		);
	
	
	
	}


	// PRIVATE... DO NOT USE DIRECTLY!
	led_private {
	
		arg column, row, state;
	
	// the following would work for a monome
	// each inheriting class should define its own way
	//	addr.sendMsg( prefix ++ "/led", column, row, state);
	
	}
	

	
	startStopBlink { arg instruction = 0;
		 
	//	arg instruction = 0, fromWhere;		// 0 = stop, 1 = play
		
//	["startStopBlink", instruction, blinkArray, fromWhere].postln;

				
		case {instruction == 0}
			{
			blinkStatus = 0;
			blinkRoutine.at(whichBlinkRoutine).stop;
			whichBlinkRoutine = (whichBlinkRoutine + 1).wrap(0, blinkRoutine.size - 1);
			}
			{instruction == 1}
			{
			blinkStatus = 1;
			blinkRoutine.at(whichBlinkRoutine).reset; 
			SystemClock.play(blinkRoutine.at(whichBlinkRoutine));
			};
	
	}
	
	// set the alphanumeric display string
	setAlpha {
	
		// this a bit of a mess at the moment
		
		arg string, cols = 16, rows = 8, start, rowOffset = 0;
		var alphaArray, colCount = 0, rowStart = 0, rowCount = 0, stringCount = 0, stopFlag = 0;
		var letterData;
		
		if(start != nil, {alphaDisplayStart = start.deepCopy}, {alphaDisplayStart = [0, 0]});
		
		rowStart = rowStart + rowOffset;

		if(string != nil,
			{
			alphaArray = Array.newClear(cols);
			alphaArray.do({
				|item, index|
				alphaArray.put(index, Array.series(rows + rowOffset, 0, 0));
			});
			
			
			while( {stopFlag == 0},
				{
				if(alphaDisplayDict.at(string.at(stringCount).toLower.asString) != nil,
					{
					letterData = alphaDisplayDict.at(string.at(stringCount).toLower.asString);
					
					if( (colCount + 1) >= cols,
						{
						colCount = 0;
						rowStart = rowStart + 4;
						if(rowStart >= rows, {stopFlag = 1});
						}
					);
					
					if( (colCount + letterData.size.div(3)) > cols,
						{
						colCount = 0;
						rowStart = rowStart + 4;
						if(rowStart >= rows, {stopFlag = 1});
						}
					);
					
					if(stopFlag != 1,
						{
					
						for( 0, letterData.size - 1,
							{
							arg c;
							alphaArray[colCount][rowStart + rowCount] = letterData[c].copy;
							rowCount = rowCount + 1;
							if(rowCount > 2, 
								{
								rowCount = 0; colCount = colCount + 1;
								if(colCount >= cols,
									{
									colCount = 0;
									rowStart = rowStart + 4;
									if(rowStart > (rows - 1), {stopFlag = 1});
									}
								);					
								
								}
							);
							
							}
						);
	
						if(colCount != 0, {colCount = colCount + 1}); rowCount = 0;
						if(colCount >= cols,
							{
							colCount = 0;
							rowStart = rowStart + 4;
							if(rowStart > (rows - 1), {stopFlag = 1});
							}
						);
						
						}
					);
							
					}
				);
				stringCount = stringCount + 1;
				if(stringCount >= string.size, {stopFlag = 1});
				}
			);
			
			alphaDisplayMap = alphaArray.deepCopy;
			
//			alphaDisplayMap.do({
//				|item, index|
//				item.postln
//			});
			
			}
		);
	
	}
	
	enableAlpha { arg overlayStyleArg = \wipe;
	
		overlayStyle = overlayStyleArg;
	
	// overlayStyle: \wipe (letters and space around them wipes any leds "underneath"
	// \exclusive: if led is 0, letter pixel is 1
	// if led is 1 or 2, letter pixel is 0
	
	// add - just gets added to everything
	
		alphaDisplayFlag = 1;
		if(alphaDisplayStart == nil, {alphaDisplayStart = [0, 0]});
		this.refreshAll;
		
	}
	
	removeAlpha {
	
		alphaDisplayFlag = 0;
		this.refreshAll;
		
	}
	
	disableAlpha {
	
//	"disableAlpha".postln;
	
		this.removeAlpha;
	
	}
	
	refreshAll {
	
		arg changeBlink = true;
		
	//	["refreahAll", changeBlink, enabled].postln;
	
		if(enabled == true,
			{
			this.refresh(0, (noOfColumns - 1), 0, (noOfRows - 1), changeBlink);
			}
		);
	
	}
	
	refreshArea {
	
		arg columnStart = 0, columnEnd = 0, rowStart = 0, rowEnd = 0, changeBlink = false;
		
		columnStart = columnStart.clip(0, (noOfColumns - 1));
		columnEnd = columnEnd.clip(0, (noOfColumns - 1));
		rowStart = rowStart.clip(0, (noOfRows - 1));
		rowEnd = rowEnd.clip(0, (noOfRows - 1));

		if(enabled == true,
			{		
			this.refresh(columnStart, columnEnd, rowStart, rowEnd, changeBlink);
			}
		);
	
	}
	
	
	// PRIVATE PRIVATE PRIVATE!!!
	refresh	{ arg columnStart, columnEnd, rowStart, rowEnd, changeBlink = true;

		var stateTemp, alphaTemp, detectIndex, responseData, buttonResponse, val, displayParam, defaultLed, butIndex, dontChangeFlag = 0;
		var letterData;
		
		var height;
		
		
	//	["REFRESH"].postln;
		
	//	["changeBlink", changeBlink].postln;
	
//	if(changeBlink == true, {"changeBlink sjddihidhs".postln});
		
		if(enabled == false, {^nil});
		
		if(alphaDisplayFlag == 1,
			{
			columnStart = 0; columnEnd = (noOfColumns - 1); rowStart = 0; rowEnd = (noOfRows - 1);
			}
		);
		
		height = rowEnd - rowStart;
		
	//	flashBag = Bag.new;

	
//		SystemClock.play(Routine({	
			for(columnStart, columnEnd,
				{
				arg column;
				
				for(rowStart, rowEnd,
					{
					arg row;	
					
					stateTemp = 0;	

					if(actionRefArray[0][column][row] != nil,
						{
						responseData = actionArray[actionDict[actionRefArray[0][column][row][0]]].deepCopy;
						buttonResponse = responseData[1];
						displayParam = responseData[2];
						defaultLed = responseData[3];
						val = responseData[4];
						butIndex = actionRefArray[0][column][row][1];
						if(buttonResponse == \exclusive,
							{
							if(val == butIndex, {stateTemp = 1 - defaultLed}, {stateTemp = defaultLed});
							},
							{
							stateTemp = val;
							}
						);
						
						if(buttonResponse == \exclusive2,
							{
							if(val == butIndex, {stateTemp = defaultLed}, {stateTemp = 0});
							}
						);

						if(buttonResponse == \slider)
							{
							if(butIndex <= val)
								{stateTemp = 1} {stateTemp = 0};
							
							};
							
						if(buttonResponse == \upSlider)
							{
							if(butIndex <= val)
								{stateTemp = 1} {stateTemp = 0};
							
							};
						
						// inclusive
						
						if(buttonResponse == \inclusive)
							{
							if(val.includes(butIndex) == true) 
								{stateTemp = 1} {stateTemp = 0};
							};
						
						if(buttonResponse == \inclusiveDown)
							{
							if(val[0].includes(butIndex) == true) 
								{stateTemp = 1} {stateTemp = 0};
							if(val[1] == butIndex)
								{
								if(flashBag.includes([column, row]) == false)
									{
									flashBag = flashBag.add([column, row])
									};
								
								};
							};

						if(buttonResponse == \exclusiveBy,
							{
							
							if(val == butIndex, 
								{
								stateTemp = 1 - defaultLed
								}, 
								{
								if(butIndex.mod(displayParam) == 0, 
									{stateTemp = 2},
									{stateTemp = defaultLed}
								);
								
								}
							);
							
							}
						);
						
						// if the function is not intended to change the led state...
						// if(stateTemp == -1, {dontChangeFlag = 1});
						
						if(buttonResponse == \noChange, {dontChangeFlag = 1});
						if(buttonResponse == \transparent, {stateTemp = stateArray[column][row]});
						
						if(buttonResponse == \fatNumber,
							{
							stateTemp = alphaDisplayDict.at(val.asString).at(butIndex);
							}
						);

						if(buttonResponse == \negButton,
							{
							stateTemp = negButtonDisplayDict.at(["+", "-"][val]).at(butIndex);
							}
						);
						
					// negButtonDisplayDict
						},
						{
						stateTemp = stateArray[column][row];
					//	["blah", stateTemp].postln;
						}
					);

					if( ((alphaDisplayFlag == 1) and: (alphaDisplayMap != nil)),
						{	
						if( (column.inclusivelyBetween(alphaDisplayStart[0], alphaDisplayStart[0] + alphaDisplayMap.size - 1))
							and:
							(row.inclusivelyBetween(alphaDisplayStart[1], alphaDisplayStart[1] + alphaDisplayMap[0].size - 1)),
								{
								alphaTemp = alphaDisplayMap[ column - alphaDisplayStart[0] ][ row - alphaDisplayStart[1] ];
								case
									{overlayStyle == \wipe}
									{stateTemp = alphaTemp}
									{overlayStyle == \exclusive}
									{
									if(alphaTemp == 1,
										{
										if(stateTemp == 0,
											{
											stateTemp = 1
											},
											{
											stateTemp = 0
											}
										);
										
										}
									);
									
									}
									{overlayStyle == \add}
									{
									if(alphaTemp == 1,
										{
										stateTemp = 1
										}
									);
									
									};
								}
						);
						
						}
					);
				
					if(dontChangeFlag != 1,
						{
						currentActualStateArray[column][row] = stateTemp;

						if(stateTemp < 2,
							{
							this.led_private(column, row, stateTemp);
							
							if(blinkArray.size > 0,
								{
								detectIndex = blinkArray.detectIndex({ arg item, i; item == [column, row] });
								if(detectIndex != nil, {blinkArray.removeAt(detectIndex)});
								}
							);
							
							},
							{
							detectIndex = blinkArray.detectIndex({ arg item, i; item == [column, row] });
							if(detectIndex == nil, {blinkArray = blinkArray.add([column, row])});
							}
						);
						
						}
					);
					
					}
				);

				}
			);
			

			
//		}));

//["changeBlink", changeBlink, flashBag.size].postln;
		if(changeBlink == true,
			{	
		//	"changeBlink is truw".postln;
	//	["bliii", (blinkArray.size > 0), (flashBag.size > 0)].postln;
			if( (blinkArray.size > 0) or: (flashBag.size > 0) )
				{
			//	if(blinkStatus != 1, {this.startStopBlink(1) });
				this.startStopBlink(0, \refresh1); this.startStopBlink(1, \refresh2);
				}
				{
				this.startStopBlink(0, \refresh3);
				};
			
			}
		);
		
	}
	
	
	clearAll {
	
//	"clear all".postln;

		this.startStopBlink(0);

		if(enabled == false, {^nil});
	
		for(0, (noOfColumns - 1),
			{
			arg column;
			for(0, (noOfRows - 1),
				{
				arg row;	
				this.led_private(column, row, 0);
				}
			);
			
			}
		);
	
	}
	
	registerFunction {
	
		arg 
			key,
			layer = 0,
			column = 0, row = 0, width = 1, height = 1, 
			defaultLed = 0, actionType = \down,
			buttonResponse = \static, responseParam = 0, displayParam = 0,
			initialValue = 0,
			geo = \scan, direction = 1,
			function, refreshFlag = 1;
			
		var actionIndex, val, butIndex, geoIndex;
		
		// actionType: \down, \up, \downUp, \none
		
		// noChange, static, toggle, scroll, exclusive
		
		// \noChange - registers key press and executes function, but does not change led status
		// \transparent - registers key press and executes function, and does not change display at all-
		// leds can be set manually and they will be visible (which is usually not the case if a function covers that area)
		// \static - area is static defaultLed. the value will always be 0.
		// \toggle - area toggles between defaultLed and 1 - defaultLed, same with value
		// \scroll - scrolls through permissable values, starting with responseParam
		
		// \exclusive - area is defaultLed except for one button, which is of value responseParam
		// value will be the number of the button selected - a single number
		
		// \fatNumber - 3 x 3 number the value of which can be changed by pressing either to the left (minus) or right (plus)
		// WIDTH AND HEIGHT ARE IGNORED this is always 3x3
		
		// (in future - thin number? why not. it wouldn't be very readable though)
		
		key = key ? ("func" ++ 3000.rand.asString);
		function = function ? {1};
		
		column = column.clip(0, noOfColumns - 1); row = row.clip(0, noOfRows - 1);

				
		if( (column + width) > noOfColumns,
			{
			width = noOfColumns - column;
			}
		);
		
		if( (row + height) > noOfRows,
			{
			height = noOfRows - row;
			}
		);

		if(buttonResponse == \fatNumber,
			{
			width = 3; height = 3;
			geo = \scan; direction = -1;
			}
		);
				
		case
			{geo == \scan}
			{
			geoIndex = C2_Geoindex.scan(width, height, direction);
			}
			{geo == \noteScan}
			{
			// reverses octaves but not individual notes:
			// octaves go from top to bottom not the other way around
			geoIndex = C2_Geoindex.scan(width, height, direction);
			geoIndex = geoIndex.clump(12).reverse.flatten(1);
			}
			{geo == \curl}
			{
			geoIndex = C2_Geoindex.curl(width, height, direction);
			};
			
//"geoindex dgdfd".postln;
//geoIndex.postcs;

		geoIndex.do({
			|item, index|
			var dontFlag = 0;
			if( (item[0] + column) > noOfColumns, {dontFlag = 1});
			if( (item[1] + row) > noOfRows, {dontFlag = 1});
			if(dontFlag == 0, {actionRefArray[layer][item[0] + column][item[1] + row] = [key, index + initialValue]});
			
		});
		
//		"bbnb".postln;
//		actionRefArray[layer].postcs;

		case
			{buttonResponse == \noChange}
			{val = 0}
			{buttonResponse == \transparent}
			{val = 0}
			{buttonResponse == \static}
			{val = defaultLed}
			{buttonResponse == \toggle}
			{val = defaultLed}
			{buttonResponse == \downToggle}
			{val = defaultLed}
			{buttonResponse == \scroll}
			{val = responseParam}
			{buttonResponse == \inclusive}
			{val = Bag.new}
			{buttonResponse == \inclusiveDown}
			{val = [Bag.new, nil]}
			{buttonResponse == \exclusive}
			{val = responseParam}
			{buttonResponse == \exclusive2}
			{val = responseParam}
			{buttonResponse == \slider}
			{val = responseParam}
			{buttonResponse == \upSlider}
			{val = responseParam}
			{buttonResponse == \exclusiveBy}
			{val = responseParam}
			{buttonResponse == \add}
			{val = responseParam}
			{buttonResponse == \fatNumber}
			{val = 0}
			{buttonResponse == \negButton}
			{val = 0};
		
		// negButton
		actionArray = actionArray.add([actionType, buttonResponse, displayParam, defaultLed, val, function, column, row, width, height]);
		actionIndex = actionArray.size - 1;
		
		actionDict.put(key, actionIndex);	
		
		if( (enabled == true) and: (refreshFlag == 1), {this.refreshAll});
		
	}
	
	removeFunctionByKey {
	// get the layer right or you're in trouble
		arg key, layer = 0, refreshFlag = 1;
		var responseData, column, row, width, height;
		
		if(actionDict[key] != nil,
			{
			responseData = actionArray[actionDict[key]].deepCopy;
			
			column = responseData[6]; 
			row = responseData[7]; 
			width = responseData[8]; 
			height = responseData[9];
			
			actionDict[key] = 0;
			
			for(column, (column + width - 1),
				{
				arg colCount;
				for(row, (row + height - 1),
					{
					arg rowCount;
					actionRefArray[layer][colCount][rowCount] = nil;
					}
				);
				
				}
			);
			
			}
		);
		if( (enabled == true) and: (refreshFlag == 1), {this.refreshAll});
		
		// actionArray = actionArray.add([actionType, buttonResponse, defaultLed, val, function, column, row, width, height]);
		// actionRefArray
	
	}
	
	getValueByKey {
	
		arg key;
		var responseData, val;
		
	//	responseData = actionArray[actionDict[actionRefArray[column][row][0]]].deepCopy;
	
		if(actionDict[key] != nil,
			{
			responseData = actionArray[actionDict[key]].deepCopy;
			val = responseData[4];
			}
		);
		
		^val;
	
	}
	
	setValueByKey {
	
		arg key, val, selectVal;
		var responseData, buttonResponse, column, row, width, height;
		
//		["setValueByKey", key, val, selectVal].postln;

		responseData = actionArray[actionDict[key]].deepCopy;
		buttonResponse = responseData[1];
		
		column = responseData[6]; 
		row = responseData[7]; 
		width = responseData[8]; 
		height = responseData[9];
		
//		if(enabled == true)
//			{	
			if(actionDict[key] != nil)
				{
				if(buttonResponse == \inclusiveDown)
					{
				//	"gtf".postln;
					if(val != nil) {actionArray[actionDict[key]][4][0] = val};
					if(selectVal != nil) 
						{
//						if(actionArray[actionDict[key]][4][1] != nil)
//							{
//							flashBag.remove([column, row]);
//							};

// aah! i can't get the above to work properly, so for now, only 1 flashbag per page hmmm

						flashBag = Bag.new;
						actionArray[actionDict[key]][4][1] = selectVal
						};
				//	["efef", selectVal, actionArray[actionDict[key]][4]].postln;
					this.refreshArea(column, column + width - 1, row, row + height - 1, true);
					}
					{
					if(val != nil)
						{
						actionArray[actionDict[key]][4] = val; 
						this.refreshArea(column, column + width - 1, row, row + height - 1);
						};
					};

				};
			
//			};
		
			
	}
	
	// these are names get/set state -> Array <- as in future
	// it might be good to be able to set / get by a list of values
	// or by a set of 8 bit values etc.
	
	getStateArray {
	
		^stateArray;
	
	}
	
	setStateArray {
	
		arg stateArrayArg;
		
		// for the moment, only accepts input of the same dimensions as currently used monome
		// maybe in future, this will clip or expand input as required
		
		if( (stateArrayArg.size == noOfColumns) and: (stateArrayArg[0].size == noOfRows),
			{
			stateArray = stateArrayArg.deepCopy;
			}
		);
	
	}
	
	clearPressBag {
	
		pressBag = Bag.new;
		pressTime = Array.fill2D( noOfColumns, noOfRows, 0 );
		flashBag = Bag.new;
		
	}
	
	handleResponse {
	
		arg column, row, status;
		var doFlag = 0;
		var elapsedTime;
		
		if(enabled == false, {^nil});
		
		// status: 1 for press down, 0 for up
		
		// thisPressBag is going to tell you all the OTHER buttons that are being pressed at the same time
		// - what their indices are, and at what Process.elapsedTime they were pressed
		
		
		case 
			{status == 1}
			{
			pressTime[column][row] = Process.elapsedTime;
			if(pressBag.includes([column, row]) == false,
				{
				pressBag.add([column, row]);
				this.private_respond(column, row, status);
				}
			);
			
			}
			{status == 0}
			{
			elapsedTime = Process.elapsedTime;
			
			SystemClock.play(Routine({
				doubleTriggerBuffer.wait;
				if( pressTime[column][row].inclusivelyBetween(elapsedTime, doubleTriggerBuffer + elapsedTime) == false,
					{
					pressBag.remove([column, row]);
					this.private_respond(column, row, status);
					}
				);
			
			}));
			
			};

	}
	
	private_respond {
	
		arg column, row, status;
		var doFlag = 0;
		var responseData, buttonResponse, actionType, func, val, butIndex, areaColumn, areaRow, areaWidth, areaHeight;
		var thisPressBag, refreshFlag;
		
		thisPressBag = Bag.new;
		pressBag.do({
			|item, index|
			if( item != [column, row],
				{
				thisPressBag.add([ item, pressTime.at(item[0]).at(item[1]) ]);
				}
			);
		
		});
	
	for(0, 1, 
		{
		arg layer;
					
		if( actionRefArray[layer][column][row] != nil,
			{
			responseData = actionArray[actionDict[actionRefArray[layer][column][row][0]]].deepCopy;
			butIndex = actionRefArray[layer][column][row][1];
			actionType = responseData[0];
			buttonResponse = responseData[1];
			val = responseData[4];
			func = responseData[5];
			
			// find the extent of the action area that the pressed led is located in
			areaColumn = responseData[6]; areaRow = responseData[7]; areaWidth = responseData[8]; areaHeight = responseData[9];
			
			// press down: status == 1
			if( (status == 1) and: ((actionType == \down) or: (actionType == \downUp)),
				{
				doFlag = 1
				}
			);
			
			// lift up: status == 0
			if( (status == 0) and: ((actionType == \up) or: (actionType == \downUp)),
				{
				doFlag = 1
				}
			);
			
			if(doFlag == 1,
				{
				case
					{buttonResponse == \noChange}
					{val = butIndex; refreshFlag = 0}
					{buttonResponse == \transparent}
					{val = butIndex; refreshFlag = 0}
					{buttonResponse == \toggle}
					{val = 1 - val; refreshFlag = 1}
					{buttonResponse == \downToggle}
					{ if(status == 1, {val = 1 - val; refreshFlag = 1}) }
					{buttonResponse == \scroll}
					{val = (val + 1).wrap(0, noOfStates - 1); refreshFlag = 1 }
					{buttonResponse == \inclusive}
					{
					if(val.includes(butIndex) == false)
						{val = val.add(butIndex); refreshFlag = 1}
						{val.remove(butIndex); refreshFlag = 1};
					}
					{buttonResponse == \inclusiveDown}
					{
					if(status == 1)
						{
						if(val[0].includes(butIndex) == false)
							{val[0] = val[0].add(butIndex); refreshFlag = 1}
							{val[0].remove(butIndex); refreshFlag = 1};
						};
					}
					{buttonResponse == \exclusive}
					{val = butIndex; refreshFlag = 1}
					{buttonResponse == \exclusive2}
					{val = butIndex; refreshFlag = 1}
					{buttonResponse == \slider}
					{val = butIndex; refreshFlag = 1}
					{buttonResponse == \upSlider}
					{val = butIndex; refreshFlag = 1}
					{buttonResponse == \exclusiveBy}
					{val = butIndex; refreshFlag = 1}
					{buttonResponse == \fatNumber}
					{
					case
						{butIndex.inclusivelyBetween(0, 2)}
						{val = (val - 1).wrap(0, 9); refreshFlag = 1}
						{butIndex.inclusivelyBetween(3, 5)}
						{val = 0; refreshFlag = 1}
						{butIndex.inclusivelyBetween(6, 8)}
						{val = (val + 1).wrap(0, 9); refreshFlag = 1};
					
					
					
					}
					{buttonResponse == \negButton}
					{
					val = (val + 1).wrap(0, 1); refreshFlag = 1
					};
					
					actionArray[actionDict[actionRefArray[layer][column][row][0]]][4] = val;
					
					func.value(val, status, thisPressBag);
					if( (refreshFlag == 1) and: (layer == 0),
						{
						this.refreshArea(areaColumn, areaColumn + areaWidth - 1, areaRow, areaRow + areaHeight - 1, "private_respond");
						}
					);
				}
			);
				
			}
		);
		
		}
	);	
		
		allAction.value(status, [column, row], thisPressBag);
	
	
	}
	
	prefix_ {
	
		arg prefixArg;
		prefix = prefixArg;
		
	}
	
	hostname_ {
		 
		arg hostnameArg;
		hostname = hostnameArg;
		addr = NetAddr(hostname, port);
	
	}
	
	listenPort_ {
	
		arg portArg;
		port = portArg;
		addr = NetAddr(hostname, port);
	
	}
	
	idNo_ {
	
		arg idNoArg;
		idNo = idNoArg;
	
	}
	

	
	


}








